import numpy as np

def mask_variance(data, background_correction=True):
    import calsipro.mask
    import calsipro.util
    assert len(data.shape) == 3
    if background_correction:
        import calsipro.background
        signal = calsipro.util.calculate_average_signal(data)
        background = calsipro.background.estimate_baseline(signal)
        data = calsipro.background.remove_background(data, background)
    mask = calsipro.mask.calculate_mask_variance(data)
    assert len(mask.shape) == 2
    return mask

def peaks(data, mask):
    import calsipro.util
    import calsipro.background
    import calsipro.peak_calling.statistics
    assert len(data.shape) == 3
    assert len(mask.shape) == 2
    signal = calsipro.util.calculate_average_signal(data, mask)

    if signal.size == 0:
        return []

    background = calsipro.background.estimate_baseline(signal)
    signal = calsipro.background.remove_background(signal, background)

    peaks = calsipro.peak_calling.statistics.call_peaks(signal)
    return peaks

def estimate_size(data):
    import numpy as snp
    import sys
    try:
        from skimage.measure import regionprops
    except ModuleNotFoundError:
        print("estimate size requires the scikit-image package.")
        sys.exit(-1)
    assert len(data.shape) == 2
    assert data.dtype == np.bool
    data = data.astype(np.uint8)
    regions = regionprops(data)

    assert len(regions) == 1
    organoid = regions[0]
    return organoid.axis_major_length



import numpy as np


def moving_average(a, n=3):
    if n == 0:
        return a
    ret = np.cumsum(a, axis=2, dtype=float)
    ret[:, :, n:] = ret[:, :, n:] - ret[:, :, :-n]
    return ret[:, :, n - 1:] / n


def max_projection(data):
    data = data[1:, :, :]
    data = np.max(data, axis=0)
    return data


def normalize(data):
    m, mm = np.min(data), np.max(data)
    data = (data - m) / (mm - m)
    return data


def normalize_to_dtype(data, dtype=np.uint8):
    m, mm = np.min(data), np.max(data)
    data = (data - m) / (mm - m)
    data = np.iinfo(np.uint8).max * data
    data = data.astype(dtype)
    return data


def calculate_mask(t, th=0.25, raw=False, labelling=True, larger=True):
    """Find the biggest connected component with t > th"""
    if not raw:
        t = np.max(t, axis=0)
        t = normalize(t)
    if larger:
        mask = t >= th
    else:
        mask = t <= th
    if not labelling:
        return mask
    import scipy.ndimage
    image = np.ones(mask.shape)
    image[~mask] = 0
    image[mask] = 1
    label, count = scipy.ndimage.label(image)
    if count == 1:
        return mask
    else:
        sizes = []
        for k in range(1, count+1):
            size = np.sum(label[mask] == k)
            sizes.append(size)
        if len(sizes) > 0:
            biggest = np.argmax(sizes)+1
        else:
            biggest = 1
        return label == biggest


def calculate_average_signal(data, mask=None):
    if mask is not None:
        signal = np.mean(data[:, mask], axis=(1))
    else:
        signal = np.mean(data, axis=(1,2))
    return signal



import numpy as np
from sklearn.linear_model import LinearRegression
from scipy.sparse import csc_matrix, eye, diags
from scipy.sparse.linalg import spsolve


def remove_background(signal, background):
    shape = list(background.shape) + [1] * (len(signal.shape) - len(background.shape))
    return signal - np.reshape(background, shape=shape)


def estimate_baseline(signal):
    corrected_signal = arPLS_fitting(signal)
    baseline = signal - corrected_signal
    return baseline


def arPLS_fitting(input_array,lam = 100, repeat = 15):
    orig_array=np.array(input_array) 
    size=orig_array.shape[0]
    w=np.ones(size)
    for i in range(1,repeat+1):
        corrected = background_estimator_aPLS(orig_array, w, lam)
        d=orig_array-corrected
        dssn=np.abs(d[d<0].sum())
        if (dssn < 0.001*(abs(orig_array)).sum() or dssn == 0.0 or i==repeat):
            if(i==repeat):
                print('baseline cannot be determined..')
            break
        w[d>=0]=0 # d>0 -> part of a peak, ignore it
        w[d<0]=np.exp(i*np.abs(d[d<0])/dssn)
        w[0]=np.exp(i*(d[d<0]).max()/dssn) 
        w[-1]=w[0]
    return orig_array-corrected


def background_estimator_aPLS(input_array,w,lam):
    array=np.matrix(input_array)
    size=array.size
    # i=np.arange(0,m)
    E=eye(size,format='csc')
    D=E[1:]-E[:-1] # numpy.diff() errors with sparse matrix .. might give errors
    W=diags(w,0,shape=(size,size))
    A=csc_matrix(W+(lam*D.T*D))
    B=csc_matrix(W*array.T)
    background=spsolve(A,B)
    return np.array(background)



from .variance import calculate_mask_variance
from .reach import calculate_mask_reach
from .intensity import calculate_mask_intensity
import numpy as np


def shrink_mask(mask, n):
    for i in range(n):
        d1 = np.diff(mask, axis=0)
        d2 = np.diff(mask, axis=1)
        mask[:-1, :][d1 != 0] = 0
        mask[1:, :][d1 != 0] = 0
        mask[:, 1:][d2 != 0] = 0
        mask[:, :-1][d2 != 0] = 0
    return mask



import numpy as np
import polars as pl
import scipy.ndimage


def time_analysis(t, intensity_cutoff=0.5):
    t = t - np.min(t, axis=0).reshape((1, t.shape[1], t.shape[2]))
    t = t / np.max(t, axis=0).reshape((1, t.shape[1], t.shape[2]))

    time = np.argmax(t >= intensity_cutoff, axis=0)
    return time


def calculate_speed(time, mask):
    """calculates speed in pixels/frame"""
    m, mm = np.min(time[mask]), np.max(time[mask])

    ori_pos = np.array(find_ori_cluster(time, mask, as_index=True)).reshape((2, 1))

    timepoints = list(range(m+1, mm+1))
    locations = [np.stack([x, y]) for x, y in find_times(time, mask, timepoints, as_index=True)]

    dts = [0]
    speeds = [-1]
    ns = [np.sum(time == m)]
    total_speed = 0
    total_ns = 0
    for t, l in zip(timepoints, locations):
        dists = np.sqrt(np.sum((l - ori_pos)**2, axis=0))
        dt = t-m
        dl = np.sum(dists)
        n = dists.shape[0]
        if n == 0:
            continue
        total_speed += dl/dt
        total_ns += n
        speeds.append(dl/(n*dt))
        ns.append(n)
        dts.append(dt)
    if total_ns == 0:
        total_ns = 1

    r = (pl.DataFrame({'time': np.array(dts, dtype=np.int64),
                       'speed': np.array(speeds, dtype=np.float64),
                       'n': np.array(ns, dtype=np.int64)}),
         total_speed/total_ns)
    return r


def push_low_pixels(time, mask, threshold=30):
    m = np.min(time[mask])
    mm = np.max(time[mask])

    time[~mask] = mm+1

    for i in range(m, mm+1):
        if np.sum(time == i) < threshold:
            time[time == i] = i+1
            m = m+1
        else:
            break

    for i in range(m, mm+1)[::-1]:
        if np.sum(time == i) < threshold:
            time[time == i] = i-1
            mm = mm-1
        else:
            break
    time[~mask] = np.max(time)
    return time


def find_times(data, mask, times, as_index=True):
    data = data.copy()
    data[~mask] = np.min(data)-1
    if as_index:
        return [(data == time).nonzero() for time in times]
    else:
        return [(data == time) for time in times]


def find_ori_cluster(data, mask, as_index=False):
    cluster_mask = np.zeros(data.shape, dtype=np.bool_)
    ori_time = np.min(data[mask])
    cluster_mask[data == ori_time] = 1
    cluster_mask[~mask] = 0
    label, count = scipy.ndimage.label(cluster_mask, scipy.ndimage.generate_binary_structure(2, 2))

    if count > 1:
        sizes = []
        for k in range(1, count+1):
            size = np.sum(label[mask] == k)
            sizes.append(size)
        biggest = np.argmax(sizes)+1
        cluster_mask = (label == biggest)

    if as_index:
        xs, ys = cluster_mask.nonzero()
        return np.array((np.average(xs), np.average(ys))).reshape((2, 1))
    else:
        return cluster_mask



import numpy as np
import pywt


def _bootstrap_baseline(time, y, level_offset=2):
    # de-biasing with wavelet
    wav = 'bior2.2'

    max_level = pywt.dwt_max_level(len(time), wav)
    idx = max(max_level-level_offset, 1)
    coeffs = pywt.wavedec(y, wav, level=idx+1)

    r = coeffs[:1] + [np.zeros(len(c)) for c in coeffs[1:]]
    s = pywt.waverec(r, wav)
    if len(s) > len(y):
        s = s[:len(y)]
    if not len(s) == len(y):
        assert len(s) == len(y)
    return s


def _find_peak_edges(peak_detected):
    i = 0
    peaks = []
    while i < len(peak_detected):
        idx_start = np.argmax(peak_detected[i:])
        if (idx_start == 0) and not peak_detected[i]:
            break
        idx_end = np.argmin(peak_detected[i+idx_start:])
        if idx_end == 0:
            idx_end = len(peak_detected[i+idx_start:])
        peaks.append((i+idx_start, min(i+idx_start+idx_end, len(peak_detected))))
        i = i + idx_start+idx_end+1
    return peaks


def _join_close_peaks(time, signal, peak_detected, peaks, N=1):
    peak_detected = peak_detected.copy()
    for p1, p2 in zip(peaks, peaks[1:]):
        p1_start, p1_end = p1
        p2_start, p2_end = p2

        if p2_start - p1_end <= N:
            peak_detected[p1_end:p2_start] = True
    return peak_detected

def _push_peak_ends(time, signal, peak_detected):
    pd = peak_detected[0]
    for i in range(1, len(time)):
        if pd and not peak_detected[i]:
            if signal[i-1] > signal[i]:
                peak_detected[i] = True
        pd = peak_detected[i]
    return peak_detected


def _call_peaks(time, signal, baseline=None, peak_detected=None, sigma=3):

    # bootstrap

    if baseline is None:
        baseline = _bootstrap_baseline(time, signal)

    if peak_detected is None:
        peak_detected = bootstrap_peaks(time, signal, baseline)

    #peak_detected = _push_peak_ends(time, signal, peak_detected)
    peaks = _find_peak_edges(peak_detected)
    peak_detected = _join_close_peaks(time, signal, peak_detected, peaks)
    peaks = _find_peak_edges(peak_detected)

    is_baseline = ~peak_detected

    # refining
    s = signal
    b = np.zeros(len(time))
    bp = np.zeros(len(time))

    d = np.zeros(len(time))
    dp = np.zeros(len(time))

    v = np.var((signal-baseline)[is_baseline])

    sbp = 1/v
    sdp = 1/np.var(np.diff(signal-baseline)[is_baseline[:-1]])

    N = 10
    p = (N-1)/N

    peak_N = 100
    peak_p = (peak_N-1)/peak_N

    b[0] = baseline[0]
    d[0] = (baseline[-1] - baseline[0]) / len(baseline)
    if d[0] > 0:
        d[0] = 0

    bp[0] = 5*sbp
    dp[0] = 5*sdp

    in_peak = False
    peak_start_b = 0

    for i in range(1, len(time)-1):
        drift_estimate = signal[i+1] - signal[i]
        b_estimate = b[i-1]+d[i-1]

        signal_bigger_than_peak_start = in_peak and signal[i] >= b[i-1]+d[i-1]
        baseline_deviates_significant = (np.abs(signal[i]-b[i-1]+d[i-1]) > 3*np.sqrt(1/bp[i-1]))
        drift_deviates_significant = ((signal[i] > b[i-1]+d[i-1]) and ((np.abs(drift_estimate-d[i-1]) > 3*np.sqrt(1/dp[i-1]))))

        # looks like a peak
        if signal_bigger_than_peak_start or baseline_deviates_significant or drift_deviates_significant:
            if not in_peak:
                in_peak = True
                peak_start_b = signal[i]
            peak_detected[i] = True
            dp[i] = dp[i-1] * peak_p
            d[i] = d[i-1]
            bp[i] = bp[i-1] * peak_p
            b[i] = b[i-1]+d[i-1]
        else:
            in_peak = False
            peak_detected[i] = False
            dp[i] = dp[i-1]*p + sdp
            d[i] = min(0, (drift_estimate*sdp + d[i-1]*dp[i-1]*p) / dp[i])

            bp[i] = bp[i-1]*p + sbp
            b[i] = ((b[i-1] + d[i-1])*bp[i-1]*p + s[i] * sbp) / bp[i]

    # no drift estimage for last value possible
    i = -1
    b_estimate = b[i-1]+d[i-1]

    # looks like a peak
    if (np.abs(signal[i]-b[i-1]+d[i-1]) > 3*np.sqrt(1/bp[i-1])) or (drift_estimate-d[i-1] > 3*np.sqrt(1/dp[i-1])):
        peak_detected[i] = True
        dp[i] = dp[i] * p
        d[i] = d[i-1]
        bp[i] = bp[i-1]
        b[i] = b[i-1]+d[i-1]
    else:
        peak_detected[i] = False
        dp[i] = dp[i-1]*p + sdp
        wd = min(sdp, 1 / (drift_estimate - d[i-1]+sdp/100)**2)
        d[i] = min(0, (drift_estimate*wd + d[i-1]*dp[i-1]*p) / (dp[i-1]*p + wd))

        bp[i] = bp[i-1]*p + sbp
        wb = 1 / (s[i] - b[i-1])**2
        b[i] = (s[i]*wb + (b[i-1] + d[i-1]) * bp[i-1] ) / (wb+bp[i-1])

    #peak_detected = _push_peak_ends(time, signal-b, peak_detected)
    peaks = _find_peak_edges(peak_detected)
    peak_detected = _join_close_peaks(time, signal, peak_detected, peaks)

    return peak_detected, b, bp, d, dp


def call_peaks(time, signal, threshold=3):
    peak_detected, b, bpp, d, dp = _call_peaks(time, signal)
    peaks = _find_peak_edges(peak_detected)

    if len(peaks) == 0:
        return peaks, b

    average_time = int(np.mean([peak[1] - peak[0] for peak in peaks]))
    change = np.lib.stride_tricks.sliding_window_view(b, average_time)
    average_change = np.mean(np.abs(change[:, -1] - change[:, 0]))

    normalised_signal = signal - b
    std = np.std(normalised_signal[~peak_detected])
    peaks = [peak for peak in peaks if np.max(normalised_signal[peak[0]:peak[1]]) > average_change*0.25]
    peaks = [peak for peak in peaks if peak[1]-peak[0] < len(signal)*0.5]
    peaks = [peak for peak in peaks if peak[1] > average_time and peak[0] < len(signal)-average_time]
    peaks = [peak for peak in peaks if peak[1] <= len(signal)-4 and peak[0] >= 4]

    return peaks, b


def bootstrap_peaks(time, measurement, baseline, sigma=1.5):
    s = measurement - baseline
    std = np.std(s)
    return s / std >= sigma

def call_peaks_simple(time, signal, threshold=1.5):
    baseline = _bootstrap_baseline(time, signal, 3)
    s = np.std(baseline)
    peak_detected = (signal-baseline) > threshold*s
    peaks = _find_peak_edges(peak_detected)
    return peaks, baseline



import calsipro.separate
import calsipro.util

def calculate_mask_intensity(data):
    assert len(data.shape) == 2
    threshold = calsipro.separate.calculate_threshold(data)
    mask = calsipro.util.calculate_mask(data, th=threshold, raw=True)
    return mask



import click
import sys

@click.group()
def cli():
    pass

@cli.command()
@click.argument("path")
def scenes(path):
    import calsipro.io
    for scene in calsipro.io.show_scenes(path):
        print(scene)

@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("out", type=click.Path(writable=True, readable=False))
@click.option("--frame", default=0)
def show(path, scene, out, frame):
    import calsipro.io
    data = calsipro.io.read_data(path, scene)[0, :, :]
    data = calsipro.util.normalize_to_dtype(data)
    calsipro.io.write_data(data, out)

@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("out", type=click.Path(writable=True, readable=False))
@click.option("--width", default=100)
@click.option("--height", default=100)
@click.option("--background-correction/--no-background-correction", default=True)
def video(path, scene, out, width, height, background_correction):
    import calsipro.io
    if out.endswith(".webm"):
        encoder = "vp9"
    elif out.endswith(".mp4"):
        encoder = "h264"
    else:
        print("Only .webm or .mp4 output files are supported")
        sys.exit(-1)


    image = calsipro.io.read_image(path)
    data = calsipro.io.read_data(image, scene)

    metadata = image.standard_metadata
    duration = metadata.total_time_duration.total_seconds()
    frames = metadata.image_size_t

    if background_correction:
        import calsipro.background
        signal = calsipro.util.calculate_average_signal(data)
        background = calsipro.background.estimate_baseline(signal)
        data = calsipro.background.remove_background(data, background)

    data = calsipro.util.normalize_to_dtype(data)
    fps = frames / duration

    bytes = calsipro.io.encode_video(data, width=width, height=height, fps=fps, encoder=encoder)
    with open(out, mode="wb") as f:
        f.write(bytes)


@cli.command(name="max")
@click.argument("path")
@click.argument("scene")
@click.argument("out")
def max_(path, scene, out):
    import calsipro.io
    import calsipro.library
    data = calsipro.io.read_data(path, scene)
    data = calsipro.util.max_projection(data)
    data = calsipro.util.normalize_to_dtype(data)
    calsipro.io.write_data(data, out)


@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("out")
@click.option("--algorithm", type=click.Choice(["variance", "reach", "intensity"]), default="variance")
@click.option("--shrink", default=5, help="Shrinks the mask from the outside by that many pixels")
@click.option("--background-correction/--no-background-correction", default=True)
def mask(path, scene, out, algorithm, background_correction, shrink):
    import calsipro.io
    import calsipro.library
    import calsipro.mask
    image = calsipro.io.read_image(path)
    data = calsipro.io.read_data(image, scene)
    if algorithm == "variance":
        mask = calsipro.library.mask_variance(data, background_correction=background_correction)
    elif algorithm == "reach":
        mask = calsipro.mask.calculate_mask_reach(data[0, :, :])
    elif algorithm == "intensity":
        mask = calsipro.mask.calculate_mask_intensity(data[0, :, :])
    if shrink > 0:
        mask = calsipro.mask.shrink_mask(mask, shrink)
    calsipro.io.write_data(mask, out)


@cli.command()
@click.argument("path")
@click.argument("scene")
@click.option("--mask")
def size(path, scene, mask):
    import calsipro.library
    import calsipro.io
    if mask is None:
        mask = calsipro.library.mask_variance(data)
    else:
        mask = calsipro.io.read_data(mask, None)[0, :, :]

    size_in_pixels = calsipro.library.estimate_size(mask)
    image = calsipro.io.read_image(path)
    size_in_μm = image.standard_metadata.pixel_size_x * size_in_pixels

    print(size_in_μm)


@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("out")
@click.option("--mask")
def peaks(path, scene, out, mask):
    import calsipro.io
    import calsipro.library
    if not (out.endswith(".parquet") or out.endswith(".csv")):
        print("out-peaks must be a parquet or csv file")
        sys.exit(-1)
    image = calsipro.io.read_image(path)
    data = calsipro.io.read_data(image, scene)

    if mask is None:
        mask = calsipro.library.mask_variance(data)
    else:
        mask = calsipro.io.read_data(mask, None)[0, :, :]

    peaks = calsipro.library.peaks(data, mask)
    table = {"peak_start": [peak[0] for peak in peaks], "peak_end": [peak[1] for peak in peaks]}
    calsipro.io.write_table(table, out)


@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("beat", type=int)
@click.argument("out")
@click.option("--peaks")
@click.option("--mask")
def beat(path, scene, out, beat, peaks, mask):
    import calsipro.io
    import calsipro.mask
    import calsipro.background
    import calsipro.library
    import calsipro.util
    import calsipro.speed
    import numpy as np
    from matplotlib import cm

    image = calsipro.io.read_image(path)
    data = calsipro.io.read_data(image, scene)

    if mask is None:
        mask = calsipro.library.mask_variance(data)
    else:
        mask = calsipro.io.read_data(mask, None)[0, :, :]

    if peaks is None:
        peaks = calsipro.library.peaks(data, mask)
    else:
        peaks = calsipro.io.read_table(peaks)
        peaks = [(start, end) for start, end in zip(peaks["peak_start"], peaks["peak_end"])]

    start, end = peaks[beat]
    data = data[start:end, :, :]

    signal = calsipro.util.calculate_average_signal(data, mask)
    background = calsipro.background.estimate_baseline(signal)
    beat_signal = calsipro.background.remove_background(data, background)

    time = calsipro.speed.time_analysis(beat_signal)

    # get a beat specific variance mask
    variance_mask = calsipro.mask.calculate_mask_variance(beat_signal)
    beat_mask = mask & variance_mask

    if beat_mask.any():
        time = calsipro.speed.push_low_pixels(time, beat_mask)

    time = calsipro.util.normalize(time)
    time = 1-time
    time = cm.inferno(time)
    time[:,:,3][~beat_mask] = 0

    time = calsipro.util.normalize_to_dtype(time)
    calsipro.io.write_data(time, out)

@cli.command()
@click.argument("path")
@click.argument("scene")
@click.argument("out")
@click.argument("flurophores", nargs=-1)
@click.option("--mask")
@click.option("--peaks")
@click.option("--fluros")
def speed(path, scene, out, flurophores, mask, peaks, fluros):
    if len(flurophores) % 2 == 1:
        print("Uneven number of flurophore arguments, need name and path for each flurophore")
        sys.exist(-1)

    from itertools import pairwise
    import calsipro.io
    import calsipro.background
    import calsipro.mask
    import calsipro.library
    import calsipro.mask
    import calsipro.speed
    from decimal import Decimal

    if fluros and flurophores:
        print("Provided flurophores via argument and via option, aborting. Provide flurophores either via argument or via option, not both.")
        sys.exit(-1)

    if fluros:
        flurophores = [f for fluro in fluros.split(";") for f in fluro.split(":")]

    image = calsipro.io.read_image(path)

    metadata = image.standard_metadata
    assert metadata.pixel_size_x == metadata.pixel_size_y
    resolution = metadata.pixel_size_x # unit: micrometer per pixel
    duration = metadata.total_time_duration.total_seconds()
    frames = metadata.image_size_t
    # speed is first calculated in pixels per frame and needs to be translated to micrometer per second
    ppf_to_μmps = resolution * frames/duration

    data = calsipro.io.read_data(image, scene)

    if mask is None:
        mask = calsipro.library.mask_variance(data)
    else:
        mask = calsipro.io.read_data(mask, None)[0, :, :]

    if peaks is None:
        peaks = calsipro.library.peaks(data, mask)
    else:
        peaks = calsipro.io.read_table(peaks)
        peaks = [(start, end) for start, end in zip(peaks["peak_start"], peaks["peak_end"])]

    fluros_names = []
    masks = []
    fluros = []
    for name, path in pairwise(flurophores):
        fluros_names.append(name)
        f_signal = calsipro.io.read_data(path, None)[0, :, :]
        f_mask = calsipro.mask.calculate_mask_intensity(f_signal)
        masks.append(f_mask)

        m, mm = np.min(f), np.max(f)
        f_norm = ((f_signal-m) / (mm-m))
        f_norm[~f_mask] = 0.0
        fluros.append(f_norm)

    unlabelled_mask = mask
    for f_mask in masks:
        unlabelled_mask = unlabelled_mask & (~f_mask)

    fluros_names.append("unlabelled")
    masks.append(unlabelled_mask)

    signal = calsipro.util.calculate_average_signal(data, mask)
    background = calsipro.background.estimate_baseline(signal)
    signal = calsipro.background.remove_background(data, background)

    results = {"beat": [], "speed": [], "origin": []}
    if len(fluros_names) > 1:
        for f_name in fluros_names:
            results[f_name + "_speed"] = []

    if mask.any():
        for peak_index, peak in enumerate(peaks):
            results["beat"].append(peak_index)
            start, end = peak
            s = signal[start:end]

            # get a beat specific variance mask
            variance_mask = calsipro.mask.calculate_mask_variance(s)
            beat_mask = mask & variance_mask

            fluros_speed = []
            ori = None
            if beat_mask.any():
                time = calsipro.speed.time_analysis(s)
                time = calsipro.speed.push_low_pixels(time, mask)
                df, speed = calsipro.speed.calculate_speed(time, mask)
                speed = ppf_to_μmps * speed

                results["speed"].append(speed)
                if len(fluros_names) > 1:
                    ori_mask = calsipro.speed.find_ori_cluster(time, mask, as_index=False)
                    f = [np.sum(fmask[ori_mask]) for fmask in fluros_mask]
                    m = np.argmax(f)
                    ori = fluros_meta[np.argmax(f)]
                    results["origin"].append(ori)

                    for fluro_name, fluro_mask in zip(fluros_names, masks):
                        if fluro_mask.any():
                            df, fluro_speed = calsipro.speed.calculate_speed(time, mask & (fluro_mask | ori_mask))
                            fluro_speed = ppf_to_μmps * fluro_speed

                            results[fluro_name + "_speed"].append(Decimal(fluro_speed).quantize(Decimal("0.01")))
                        else:
                            results[fluro_name + "_speed"].append(None)
                else:
                    results["origin"].append(None)
            else:
                results["speed"].append(None)
                results["origin"].append(None)
                if len(fluros_names) > 1:
                    for fluro_name in fluros_names:
                        results[fluro_name + "_speed"].append(None)
    calsipro.io.write_table(results, out)



import numpy as np
import logging


def calculate_threshold(data, pick=1):

    if np.min(data) == 0:
        offset = 1
        logging.debug('Calculating threshold, min of data is 0, setting offset to 1.')
    else:
        offset = 0

    freq, counts, bins = calculate_histogram(data)
    left_flood = freq.copy()
    right_flood = freq.copy()
    flood = freq.copy()

    for i in range(1, len(freq)):
        left_flood[i] = max(left_flood[i], left_flood[i-1])

    for i in list(range(0, len(freq)-1))[::-1]:
        right_flood[i] = max(right_flood[i], right_flood[i+1])

    for i in range(0, len(freq)):
        flood[i] = min(left_flood[i], right_flood[i])

    f = flood - freq
    if pick != 1:
        idxs = np.argsort(flood-freq)
        idx = idxs[-pick]
    else:
        idx = np.argmax(flood-freq)
    logging.debug(f'Calculating threshold, index is {idx}')

    rest = freq[idx:]
    low = freq[idx]
    high = np.max(rest)

    idx_offset = max(0, np.argmax(rest >= (low + (high-low)*0.10))-1)
    idx = idx + idx_offset

    logging.debug(f'Calculating threshold, index was pushed to {idx}')


    pixels = np.sum(counts[idx:]) / np.sum(counts)
    if  pixels < 0.001:
        logging.debug('Calculating threshold, threshold is lower than 0.001 percentile, invalidating threshold')
        return np.max(data)+1
    if  0.999 < pixels:
        logging.debug('Calculating threshold, threshold is higher than 0.999 percentile, invalidating threshold')
        return np.max(data)+1
    if idx == 0:
        logging.debug('Calculating threshold, threshold is 0, invalidating threshold')
        return np.max(data)+1
    threshold = np.exp(bins[idx+1])-offset
    logging.debug('Calculating threshold, final threshold is %s', threshold)
    return threshold


def calculate_histogram(data):
    N = 80
    zero_bins = 80
    while zero_bins > 4:
        freq, counts, bins = _calculate_histogram(data, N)
        zero_bins = np.sum(counts == 0)
        N = int(N/2)
    return freq, counts, bins


def _calculate_histogram(data, N):
    if data.dtype == np.uint8 or data.dtype == np.uint16:
        data = data.astype(np.int32)
    if np.min(data) == 0:
        offset = 1
    else:
        offset = 0
    try:
        counts, bins = np.histogram(np.log(data+offset), N)
    except Exception as e:
        d1 = data+offset
        d2 = np.log(d1)
        logging.error({'data+offset': d1, 'log(data+offset)': d2, 'offset': offset, 'data min': np.min(data), 'data max': np.max(data), 'data+offset min': np.min(d1), 'data+offset max': np.max(d1), 'log(data+offset min)': np.min(d2), 'log(data+offset max)': np.max(d2)})
        raise e

    for i in range(len(counts)):
        if counts[i] <= 10:
            counts[i] = 1
        else:
            break

    for i in range(1, len(counts)):
        if counts[-i] <= 10:
            counts[-i] = 1
        else:
            break
    freq = np.log(1+counts)
    return freq, counts, bins



import calsipro.separate
import calsipro.util
import numpy as np

def calculate_mask_variance(data):
    d = np.var(data, axis=0)
    threshold = calsipro.separate.calculate_threshold(d)
    mask = calsipro.util.calculate_mask(d, th=threshold, raw=True)
    return mask



from PIL import Image
import openpyxl
import numpy as np
import polars as pl
from calsipro.util import moving_average
from bioio import BioImage
import ffmpeg
import select

def read_image(img):
    if isinstance(img, BioImage):
        return img
    return BioImage(img, use_aicspylibczi=True)


def show_scenes(inp):
    img = read_image(inp)
    return img.scenes


def read_data(inp, scene=None, requested_frames=None, average=0):
    img = read_image(inp)

    if scene is None:
        scene = img.scenes[0]

    img.set_scene(scene)

    if requested_frames is None:
        frames = img.dims['T'][0]
        requested_frames = [(0, frames-average-1)]

    data = img.data
    images = []
    for frame_request in requested_frames:
        start, stop = frame_request
        if start < average:
            start_cut = average-start
        else:
            start_cut = average
        start, stop = start-start_cut, stop+average

        start = max(0, start)
        stop = max(min(frames-1, stop), start+1)

        image = data[start:stop, 0, 0, :, :]

        if average > 1:
            image = moving_average(image, n=average)
        image = image[start_cut:stop-start-average, :, :]
        images.append(image)
    if len(images) == 1:
        return images[0]
    return images


def write_table(df, path):
    df = pl.DataFrame(df)
    if path.endswith(".parquet"):
        df.write_parquet(path)
    elif path.endswith(".csv"):
        df.write_csv(path)
    else:
        raise KeyError("File Ending not supported")


def read_table(path):
    if path.endswith(".parquet"):
        return pl.read_parquet(path)
    elif path.endswith(".csv"):
        return pl.read_csv(path)
    else:
        raise KeyError("File Ending not supported")


WRITER_PACKAGES = {
    "OmeTiffWriter": "bioio-ome-tiff",
    "OmeZarrWriterV3": "bioio-ome-zarr",
    "TimeSeriesWriter": "bioio-imageio",
    "TwoDWriter": "bioio-imageio",
}


WRITERS = {
    ".ome.tiff": "OmeTiffWriter",
    ".ome.tif": "OmeTiffWriter",
    ".ome.zarr": "OmeZarrWriterV3",
    ".gif": "TimeseriesWriter",
    ".mp4": "TimeseriesWriter",
    ".mkv": "TimeseriesWriter",
    ".png": "TwoDWriter",
    ".bmp": "TwoDWriter",
    ".jpg": "TwoDWriter",
    }


def _get_writer(name):
    parts = name.split(".")
    potential_endings = []
    if len(parts) > 2:
        potential_endings.append("." + parts[-2] + "." + parts[-1])
    if len(parts) > 1:
        potential_endings.append("." + parts[-1])

    writer_name = None
    for ending in potential_endings:
        writer_name = WRITERS.get(ending)
        if writer_name:
            break

    if writer_name is None:
        raise KeyError(f"Could not identify writer for path `{name}`")

    import bioio.writers
    try:
        return getattr(bioio.writers, writer_name)
    except AttributeError:
        raise ValueError(f"Writer {writer_name} is not installed, install package {WRITER_PACKAGES[writer_name]}.")


def write_data(data, path):
    writer = _get_writer(path)
    writer.save(data, path)


def encode_video(data, width=100, height=100, fps=5, encoder="h264"):
    if encoder == "vp9":
        p = (ffmpeg
        .input('pipe:', format='rawvideo', pix_fmt='gray', s='{}x{}'.format(data.shape[-2], data.shape[-1]), r=fps)
        .output("pipe:1", format="webm", s=f"{width}x{height}", pix_fmt="yuv420p", deadline="realtime", crf="30", **{"row-mt": "1", "b:v": "2000k", "tile-columns": 3, "c:v": "libvpx-vp9"})
        .run_async(pipe_stdin=True, pipe_stdout=True, pipe_stderr=True))
    elif encoder == "h264":
        p = (ffmpeg
        .input('pipe:', format='rawvideo', pix_fmt='gray', s='{}x{}'.format(data.shape[-2], data.shape[-1]), r=fps)
        .output("pipe:1", format="mp4", movflags="frag_keyframe+empty_moov", s=f"{width}x{height}", pix_fmt="yuv420p", preset="fast", tune="zerolatency", crf="22", **{"c:v": "libx264"})
        .run_async(pipe_stdin=True, pipe_stdout=True, pipe_stderr=True))
    else:
        raise ValueError(f"Unknown encoder {encoder}, only supported are 'vp9' and 'h264'")

    b = data.tobytes()
    outs, errs = p.communicate(input=b)

    if len(outs) == 0:
        raise Exception("Video was not generated correctly\n\n"+errs.decode("utf8"))

    return outs




import numpy as np

def call_peaks(s, sigma=1.5):
    std = np.std(s)
    peaks_detected = s / std >= sigma
    peaks = _find_peak_edges(peaks_detected)
    peaks_detected = _join_close_peaks(s, peaks_detected, peaks)
    peaks = _find_peak_edges(peaks_detected)

    peaks = [push_peak(peak, s) for peak in peaks]

    return peaks

def push_peak(peak, signal):
    start, end = peak
    m = len(signal)
    while start > 0 and signal[start-1] < signal[start]:
        start -= 1
    while end < m-1 and signal[end+1] < signal[end]:
        end += 1
    return start, end

def _find_peak_edges(peak_detected):
    i = 0
    peaks = []
    while i < len(peak_detected):
        idx_start = np.argmax(peak_detected[i:])
        if (idx_start == 0) and not peak_detected[i]:
            break
        idx_end = np.argmin(peak_detected[i+idx_start:])
        if idx_end == 0:
            idx_end = len(peak_detected[i+idx_start:])
        peaks.append((i+idx_start, min(i+idx_start+idx_end, len(peak_detected))))
        i = i + idx_start+idx_end+1
    return peaks

def _join_close_peaks(signal, peak_detected, peaks, N=1):
    peak_detected = peak_detected.copy()
    for p1, p2 in zip(peaks, peaks[1:]):
        p1_start, p1_end = p1
        p2_start, p2_end = p2

        if p2_start - p1_end <= N:
            peak_detected[p1_end:p2_start] = True
    return peak_detected





import numpy as np
import numba
import heapq

import calsipro.util
import calsipro.separate



def calculate_mask_reach(data):
    assert len(data.shape) == 2
    data = data.copy()
    reach = reachability(data)
    th = calsipro.separate.calculate_threshold(reach)
    mask = calsipro.util.calculate_mask(reach, th=th, raw=True)
    return mask


def reachability(data):
    data = calsipro.util.normalize(data)
    reach = np.ones(shape=data.shape, dtype=data.dtype)
    visited = np.zeros(shape=data.shape, dtype=np.bool_)
    level = np.zeros(shape=data.shape, dtype=data.dtype)
    _reachability(data, reach, visited, level)
    return reach


@numba.njit(cache=True)
def _reach_init_point(x, y, next, reach, visited, data, level):
    heapq.heappush(next, (0.0, x, y))
    reach[x, y] = 0.0
    level[x, y] = data[x, y]


@numba.njit(cache=True)
def _reach_init(xdim, ydim, next, reach, visited, data, level):

    for y in range(10):
        _reach_init_point(0, y, next, reach, visited, data, level)
        _reach_init_point(0, ydim-y-1, next, reach, visited, data, level)
        _reach_init_point(xdim-1, y, next, reach, visited, data, level)
        _reach_init_point(xdim-1, ydim-y-1, next, reach, visited, data, level)

    for x in range(10):
        _reach_init_point(x, 0, next, reach, visited, data, level)
        _reach_init_point(xdim-x-1, 0, next, reach, visited, data, level)
        _reach_init_point(x, ydim-1, next, reach, visited, data, level)
        _reach_init_point(xdim-x-1, ydim-1, next, reach, visited, data, level)


@numba.njit(cache=True)
def _reach_check(ox, oy, x, y, xdim, ydim, next, reach, data, visited, level):
    W1 = 100
    W2 = 1
    WT = W1+W2
    if 0 <= x < xdim and 0 <= y < ydim:
        r = abs(data[x,y] - level[ox, oy])
        r = max(r, reach[ox, oy])
        reach[x, y] = min(reach[x, y], r)
        level[x, y] = (W1*level[ox, oy] + W2*data[x, y]) / WT
        heapq.heappush(next, (r, x, y))


@numba.njit(cache=True)
def _reach_visit(ox, oy, xdim, ydim, next, reach, data, visited, level):
    if visited[ox, oy]:
        return
    visited[ox, oy] = True

    _reach_check(ox, oy, ox+1, oy, xdim, ydim, next, reach, data, visited, level)
    _reach_check(ox, oy, ox-1, oy, xdim, ydim, next, reach, data, visited, level)
    _reach_check(ox, oy, ox, oy+1, xdim, ydim, next, reach, data, visited, level)
    _reach_check(ox, oy, ox, oy-1, xdim, ydim, next, reach, data, visited, level)


@numba.njit(cache=True)
def _reachability(data, reach, visited, level):
    next = [(np.float64(0.0), 1, 1) for x in range(0)]
    xdim, ydim = data.shape
    _reach_init(xdim, ydim, next, reach, visited, data, level)

    while (len(next) != 0):
        r, x, y = heapq.heappop(next)
        _reach_visit(x, y, xdim, ydim, next, reach, data, visited, level)
